var ArrayTracker = require('../index.js')
    , mocha = require('mocha')
    , expect = require('chai').expect;

describe('ArrayTracker', function () {
    var emptyArray, stringArray, simpleArray, objArray, deepArray;
    before(function () {
        tracker = new ArrayTracker();
        emptyArray = [];
        simpleArray = [1, 2, 3];
        objArray = [
            {letter: 'a', number: 1},
            {letter: 'b', number: 1},
            {letter: 'c', number: 3}
        ];
        deepArray = { test: true, deep: { deeper: { deepest: { arr: [
            {id: 1, name: 'first', age: 10},
            {id: 2, name: 'second', age: 10},
            {id: 3, name: 'third', age: 14},
            {id: 4, name: 'third', age: 10}
        ]  }}}};
    });

    describe("Initialization", function () {
        it('should exist', function () {
            expect(tracker).to.exist;
        });
        /*it('should error if not passed an array', function(){
         expect(tracker.initialize()).to.throw(Error);
         });*/
        it('should initialize an empty array', function () {
            expect(tracker.initialize(emptyArray)).to.be.true;
        });
        it('should initialize a simple array', function () {
            expect(tracker.initialize(simpleArray)).to.be.true;
        });
        it('should initialize a object array', function () {
            expect(tracker.initialize(objArray)).to.be.true;
        });
        it('should initialize a object array with a key', function () {
            expect(tracker.initialize(objArray, ['letter'])).to.be.true;
        });
        it('should initialize a object array with a deep array', function () {
            expect(tracker.initialize(deepArray, 'deep.deeper.deepest.arr')).to.be.true;
        });
        it('should initialize a object array with a deep array with a key', function () {
            expect(tracker.initialize(deepArray, 'deep.deeper.deepest.arr', ['id', 'name'])).to.be.true;
        });
    });

    describe("Tracking an empty array", function () {
        before(function () {
            tracker = new ArrayTracker();
            emptyArray = [];
            tracker.initialize(emptyArray);
        });
        it('should detect no change', function () {
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker.evaluate(emptyArray).itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(addResult).to.be.undefined;
            expect(deleteResult).to.be.undefined;
            expect(changedResult).to.be.undefined;
            expect(unchangedResult).to.equal(emptyArray);
        });
        it('should allow for different configurations of method chaining', function () {
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker.evaluate(emptyArray).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(changedResult).to.be.undefined;
            expect(unchangedResult).to.equal(emptyArray);
        });
        it('should detect an added item', function () {
            var addResult, deleteResult, unchangedResult, changedResult;
            emptyArray.push(1);
            tracker.evaluate(emptyArray).itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(addResult).to.have.length(1);
            expect(addResult).to.include(1);
            expect(changedResult).to.equal(emptyArray);
            expect(deleteResult).to.be.undefined;
            expect(unchangedResult).to.be.undefined;

        });
        it('should detect a deleted item', function () {
            var addResult, deleteResult, unchangedResult, changedResult;
            emptyArray.splice(0, 1);
            tracker.evaluate(emptyArray).itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(deleteResult).to.have.length(1);
            expect(deleteResult).to.include(1);
            expect(addResult).to.be.undefined;
            expect(unchangedResult).to.be.undefined;
            expect(changedResult).to.equal(emptyArray);

        });
    });

    describe("Tracking a simple array", function () {
        before(function () {
            tracker = new ArrayTracker();
            simpleArray = [1, 2, 3];
            tracker.initialize(simpleArray);
        });
        it('should detect no change', function () {
            var addResult, deleteResult, unchangedResult, changedResult;
            var result;
            tracker.evaluate(simpleArray).itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(unchangedResult).to.equal(simpleArray)
            expect(deleteResult).to.be.undefined;
            expect(addResult).to.be.undefined;
            expect(changedResult).to.be.undefined;
        });
        it('should detect an added item', function () {
            var addResult, deleteResult, unchangedResult, changedResult;
            simpleArray.push(4);
            tracker.evaluate(simpleArray).itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(addResult).to.have.length(1);
            expect(addResult).to.include(4);
            expect(deleteResult).to.be.undefined;
            expect(unchangedResult).to.be.undefined;
            expect(changedResult).to.equal(simpleArray);
        });
        it('should detect a deleted item', function () {
            var addResult, deleteResult, unchangedResult, changedResult;
            simpleArray.splice(0, 1);
            tracker.evaluate(simpleArray).itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });

            expect(deleteResult).to.have.length(1);
            expect(deleteResult).to.include(1);
            expect(addResult).to.be.undefined;
            expect(unchangedResult).to.be.undefined;
            expect(changedResult).to.equal(simpleArray);
        });
        it('should detect wholesale change', function () {
            var addResult, deleteResult, unchangedResult, changedResult;
            var newArr = [4, 5, 6];
            var result;
            tracker.evaluate(newArr).itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });

            expect(deleteResult).to.have.length(2);
            expect(deleteResult).to.include(2, 3);
            expect(addResult).to.have.length(2);
            expect(addResult).to.include(5, 6)
            expect(unchangedResult).to.be.undefined;
            expect(changedResult).to.equal(newArr);
        });


    });

    describe("Tracking arrays with duplicate simple types", function () {

        it('should allow duplicate items', function () {
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker = new ArrayTracker();
            simpleArray = [1, 2, 3];
            tracker.initialize(simpleArray);
            simpleArray.push(4);
            tracker.evaluate(simpleArray).itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(addResult).to.have.length(1);
            expect(addResult).to.include(4);
            expect(deleteResult).to.be.undefined;
            expect(unchangedResult).to.be.undefined;
            expect(changedResult).to.equal(simpleArray);
        });
        it('should allow multiple duplicate items', function () {
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker = new ArrayTracker();
            simpleArray = [1, 2, 3];
            tracker.initialize(simpleArray);
            simpleArray.push(3);
            simpleArray.push(3);
            tracker.evaluate(simpleArray).itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(addResult).to.have.length(2);
            expect(addResult).to.include(3);
            expect(deleteResult).to.be.undefined;
            expect(unchangedResult).to.be.undefined;
            expect(changedResult).to.equal(simpleArray);
        });
        it('should handle deleting duplicate items', function () {
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker = new ArrayTracker();
            simpleArray = [1, 2, 3, 3, 3];
            tracker.initialize(simpleArray);
            simpleArray.splice(3,2);
            tracker.evaluate(simpleArray).itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(deleteResult).to.have.length(2);
            expect(deleteResult).to.include(3);
            expect(addResult).to.be.undefined;
            expect(unchangedResult).to.be.undefined;
            expect(changedResult).to.equal(simpleArray);
        });
    });

    describe("Tracking arrays with duplicate object types", function () {
        it("should handle multiple duplicate objects",function(){
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker = new ArrayTracker();
            objArray = [
                {letter: 'a', number: 1},
                {letter: 'b', number: 1},
                {letter: 'c', number: 3}
            ];
            tracker.initialize(objArray);
            objArray.push({letter: 'c', number: 3});
            objArray.push({letter: 'c', number: 3});
            tracker.evaluate(objArray).itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(addResult).to.have.length(2);
            expect(addResult[0]).to.deep.equal({letter: 'c', number: 3});
            expect(addResult[1]).to.deep.equal({letter: 'c', number: 3});
            expect(deleteResult).to.be.undefined;
            expect(unchangedResult).to.be.undefined;
            expect(changedResult).to.equal(objArray);
        });

        it("should handle deleting multiple duplicate objects",function(){
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker = new ArrayTracker();
            objArray = [
                {letter: 'a', number: 1},
                {letter: 'b', number: 1},
                {letter: 'c', number: 3},
                {letter: 'c', number: 3},
                {letter: 'c', number: 3}
            ];
            tracker.initialize(objArray);
            objArray.splice(3,2);
            tracker.evaluate(objArray).itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(deleteResult).to.have.length(2);
            expect(deleteResult[0]).to.deep.equal({letter: 'c', number: 3});
            expect(deleteResult[1]).to.deep.equal({letter: 'c', number: 3});
            expect(addResult).to.be.undefined;
            expect(unchangedResult).to.be.undefined;
            expect(changedResult).to.equal(objArray);
        })

    });

    describe("Tracking a object array", function () {
        before(function () {
            tracker = new ArrayTracker();
            objArray = [
                {letter: 'a', number: 1},
                {letter: 'b', number: 1},
                {letter: 'c', number: 3}
            ];
            tracker.initialize(objArray);
        });
        it('should detect no change', function () {
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker.evaluate(objArray).itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;

                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(unchangedResult).to.deep.equal(objArray);
            expect(deleteResult).to.be.undefined;
            expect(addResult).to.be.undefined;
            expect(changedResult).to.be.undefined;
        });
        it('should detect an added item', function () {
            objArray.push({letter: 'd', number: 4});
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker.evaluate(objArray).itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(addResult).to.have.length(1);
            expect(addResult[0]).to.deep.equal({letter: 'd', number: 4});
            expect(deleteResult).to.be.undefined;
            expect(unchangedResult).to.be.undefined;
            expect(changedResult).to.equal(objArray);
        });
        it('should detect a deleted item', function () {
            objArray.splice(0, 1);
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker.evaluate(objArray).itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });

            expect(deleteResult).to.have.length(1);
            expect(deleteResult[0]).to.deep.equal({letter: 'a', number: 1});
            expect(addResult).to.be.undefined;
            expect(unchangedResult).to.be.undefined;
            expect(changedResult).to.equal(objArray);
        });
        it('should detect a wholesale change', function () {
            objArray = [
                {letter: 'z', number: 1},
                {letter: 'w', number: 1},
                {letter: 'c', number: 3}
            ];
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker.evaluate(objArray).itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(addResult).to.have.length(2);
            expect(deleteResult).to.have.length(2);
            expect(unchangedResult).to.be.undefined;
            expect(changedResult).to.equal(objArray);
        });
    });

    describe("Tracking a object array with a keyed Id", function () {
        before(function () {
            tracker = new ArrayTracker();
            objArray = [
                {letter: 'a', number: 1},
                {letter: 'b', number: 1},
                {letter: 'c', number: 3}
            ];
            tracker.initialize(objArray,['letter']);
        });
        it('should detect no change', function () {
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker.evaluate(objArray).itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;

                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(unchangedResult).to.deep.equal(objArray);
            expect(deleteResult).to.be.undefined;
            expect(addResult).to.be.undefined;
            expect(changedResult).to.be.undefined;
        });
        it('should allow changes to non-keyed fields', function () {
            var addResult, deleteResult, unchangedResult, changedResult;
            objArray = [
                {letter: 'a', number: 100},
                {letter: 'b', number: 100},
                {letter: 'c', number: 100}
            ];
            tracker.evaluate(objArray).itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;

                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(unchangedResult).to.deep.equal(objArray);
            expect(deleteResult).to.be.undefined;
            expect(addResult).to.be.undefined;
            expect(changedResult).to.be.undefined;
        });
        it('should detect an added item', function () {
            objArray.push({letter: 'd', number: 4});
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker.evaluate(objArray).itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(addResult).to.have.length(1);
            expect(addResult[0]).to.deep.equal({letter: 'd', number: 4});
            expect(deleteResult).to.be.undefined;
            expect(unchangedResult).to.be.undefined;
            expect(changedResult).to.equal(objArray);
        });
        it('should detect a deleted item', function () {
            objArray.splice(0, 1);
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker.evaluate(objArray).itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });

            expect(deleteResult).to.have.length(1);
            expect(deleteResult[0]).to.deep.equal({letter: 'a', number: 1});
            expect(addResult).to.be.undefined;
            expect(unchangedResult).to.be.undefined;
            expect(changedResult).to.equal(objArray);
        });
        it('should detect a wholesale change', function () {
            objArray = [
                {letter: 'z', number: 1},
                {letter: 'w', number: 1},
                {letter: 'c', number: 3}
            ];
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker.evaluate(objArray).itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(addResult).to.have.length(2);
            expect(deleteResult).to.have.length(2);
            expect(unchangedResult).to.be.undefined;
            expect(changedResult).to.equal(objArray);
        });
    });

    describe("Tracking an embedded array", function () {
        before(function () {
            tracker = new ArrayTracker();
            deepArray = { test: true, deep: { deeper: { deepest: { arr: [
                {id: 1, name: 'first', age: 10},
                {id: 2, name: 'second', age: 10},
                {id: 3, name: 'third', age: 14}
            ]  }}}};
            tracker.initialize(deepArray, 'deep.deeper.deepest.arr');
        });
        it('should detect no change', function () {
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker.evaluate(deepArray, 'deep.deeper.deepest.arr').itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(unchangedResult).to.deep.equal(deepArray);
            expect(addResult).to.be.undefined;
            expect(deleteResult).to.be.undefined;
            expect(changedResult).to.be.undefined;
        });
        it('should detect an added item', function () {
            deepArray.deep.deeper.deepest.arr.push({id: 4, name: 'third', age: 10});
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker.evaluate(deepArray, 'deep.deeper.deepest.arr').itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(addResult).to.have.length(1);
            expect(unchangedResult).to.be.undefined;
            expect(deleteResult).to.be.undefined;
            expect(changedResult).to.equal(deepArray);
        });
        it('should detect a deleted item', function () {
            deepArray.deep.deeper.deepest.arr.splice(0, 1);
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker.evaluate(deepArray, 'deep.deeper.deepest.arr').itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });

            expect(deleteResult).to.have.length(1);
            expect(addResult).to.be.undefined;
            expect(unchangedResult).to.be.undefined;
            expect(changedResult).to.equal(deepArray);
        });

    });
    describe("Tracking an embedded array that contain arrays", function () {
        before(function () {
            tracker = new ArrayTracker();
            deepArray = { test: true, deep: { deeper: { deepest: { arr: [
                {id: 1, name: 'first', age: 10, nicknames:['uno','onesy']},
                {id: 2, name: 'second', age: 10},
                {id: 3, name: 'third', age: 14}
            ]  }}}};
        });
        it('should detect no change', function () {
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker.evaluate(deepArray, 'deep.deeper.deepest.arr[0]').itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(unchangedResult).to.deep.equal(deepArray);
            expect(addResult).to.be.undefined;
            expect(deleteResult).to.be.undefined;
            expect(changedResult).to.be.undefined;
        });
    });

    describe("Tracking a deep object array with a defined key", function () {
        before(function () {
            tracker = new ArrayTracker();
            deepArray = { test: true, deep: { deeper: { deepest: { arr: [
                {id: 1, name: 'first', age: 10},
                {id: 2, name: 'second', age: 10},
                {id: 3, name: 'third', age: 14}
            ]  }}}};
            tracker.initialize(deepArray, 'deep.deeper.deepest.arr', ['id']);
        });
        it('should detect no change', function () {
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker.evaluate(deepArray, 'deep.deeper.deepest.arr').itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(unchangedResult).to.deep.equal(deepArray);
            expect(addResult).to.be.undefined;
            expect(deleteResult).to.be.undefined;
            expect(changedResult).to.be.undefined;
        });
        it('should detect no change on unkeyed value', function () {
            var addResult, deleteResult, unchangedResult, changedResult;
            deepArray.deep.deeper.deepest.arr[0].birthday = "01/01/1970";
            tracker.evaluate(deepArray, 'deep.deeper.deepest.arr').itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(unchangedResult).to.deep.equal(deepArray);
            expect(addResult).to.be.undefined;
            expect(deleteResult).to.be.undefined;
            expect(changedResult).to.equal(undefined);
        });
        it('should detect an added item', function () {
            deepArray.deep.deeper.deepest.arr.push({id: 4, name: 'third', age: 10});
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker.evaluate(deepArray, 'deep.deeper.deepest.arr').itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(addResult).to.have.length(1);
            expect(unchangedResult).to.be.undefined;
            expect(deleteResult).to.be.undefined;
            expect(changedResult).to.equal(deepArray);
        });

        it('should allow for different combinations of method chaining', function () {
            deepArray.deep.deeper.deepest.arr.push({id:5, name: 'fifth', age: 10});
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker.evaluate(deepArray, 'deep.deeper.deepest.arr').itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                deleteResult = arr;
                });
            expect(addResult).to.have.length(1);
            expect(deleteResult).to.be.undefined;
        });
        it('should detect a deleted item', function () {
            deepArray.deep.deeper.deepest.arr.splice(0, 1);
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker.evaluate(deepArray, 'deep.deeper.deepest.arr').itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });

            expect(deleteResult).to.have.length(1);
            expect(addResult).to.be.undefined;
            expect(unchangedResult).to.be.undefined;
            expect(changedResult).to.equal(deepArray);
        });
        it('should error on duplicate keys', function () {
            deepArray.deep.deeper.deepest.arr.push({id: 4, name: 'fourth', age: 10});
            var addResult, deleteResult, unchangedResult, changedResult;
            try {
                tracker.evaluate(deepArray, 'deep.deeper.deepest.arr')
            } catch (e) {
                expect(e.message).to.equal('Duplicate key');
            }

        });
    });

    describe("Tracking a deep object array with multiple defined key", function () {
        before(function () {
            tracker = new ArrayTracker();
            deepArray = { test: true, deep: { deeper: { deepest: { arr: [
                {id: 1, name: 'first', age: 10},
                {id: 2, name: 'second', age: 10},
                {id: 3, name: 'third', age: 14}
            ]  }}}};
            tracker.initialize(deepArray, 'deep.deeper.deepest.arr', ['id', 'name']);
        });
        it('should detect no change', function () {
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker.evaluate(deepArray, 'deep.deeper.deepest.arr').itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(unchangedResult).to.deep.equal(deepArray);
            expect(addResult).to.be.undefined;
            expect(deleteResult).to.be.undefined;
            expect(changedResult).to.be.undefined;
        });
        it('should detect no change on unkeyed values', function () {
            var addResult, deleteResult, unchangedResult, changedResult;
            deepArray.deep.deeper.deepest.arr[0].age = "Pork Chops";
            deepArray.deep.deeper.deepest.arr[1].age = "Apple sauce";
            tracker.evaluate(deepArray, 'deep.deeper.deepest.arr').itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(unchangedResult).to.deep.equal(deepArray);
            expect(addResult).to.be.undefined;
            expect(deleteResult).to.be.undefined;
            expect(changedResult).to.equal(undefined);
        });
        it('should detect an added item', function () {
            deepArray.deep.deeper.deepest.arr.push({id: 4, name: 'third', age: 10});
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker.evaluate(deepArray, 'deep.deeper.deepest.arr').itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });
            expect(addResult).to.have.length(1);
            expect(unchangedResult).to.be.undefined;
            expect(deleteResult).to.be.undefined;
            expect(changedResult).to.equal(deepArray);
        });
        it('should detect a deleted item', function () {
            deepArray.deep.deeper.deepest.arr.splice(0, 1);
            var addResult, deleteResult, unchangedResult, changedResult;
            tracker.evaluate(deepArray, 'deep.deeper.deepest.arr').itemsAdded(function (arr) {
                addResult = arr;
            }).itemsDeleted(function (arr) {
                    deleteResult = arr;
                }).unchanged(function (arr) {
                    unchangedResult = arr;
                }).changed(function (arr) {
                    changedResult = arr;
                });

            expect(deleteResult).to.have.length(1);
            expect(addResult).to.be.undefined;
            expect(unchangedResult).to.be.undefined;
            expect(changedResult).to.equal(deepArray);
        });
    });


});
